/*
** Module   : WidgetFactory.java
** Abstract : Provides factory APIs to create widgets using the CREATE WIDGET statement. 
**
** Copyright (c) 2013, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- ---------------------------------Description----------------------------------
** 001 CA  20130112 Created initial version.
** 002 CA  20130918 Added widget-pool support. The pool name must be validated before anything 
**                  else, and a null name means the default pool is used.
** 003 CA  20130927 Widget resource type are determined from LegacyResource constants.
** 004 CA  20131021 Mark the resources as being created dynamic.
** 005 CA  20131115 Changes related to properly instantiate dynamic resources.
** 006 VIG 20131127 Added method to check possibility to create new window
** 007 EVL 20140918 Adding support for dynamic widget.
** 008 CA  20140929 Replaced WidgetId.allocateId with the WidgetId.nextId API, which returns a
**                  WidgetId instance.
** 009 VIG 20141104 Added MENU widget support.
*/

package com.goldencode.p2j.ui;

import java.lang.reflect.*;
import java.util.*;
import com.goldencode.p2j.util.*;

/**
 * Provides APIs to create the following widgets:
 * <table>
 * <tr>
 *    <td>Widget Type</td>
 *    <td>API</td>
 * <tr>
 * <tr>
 *    <td>CREATE VALUE(...)</td>
 *    <td>{@link #create}</td>
 * </tr>
 * <tr>
 *    <td>CREATE BUTTON</td>
 *    <td>{@link #createButton}</td>
 * </tr>
 * <tr>
 *    <td>CREATE COMBO-BOX</td>
 *    <td>{@link #createComboBox}</td>
 * </tr>
 * <tr>
 *    <td>CREATE CONTROL-FRAME</td>
 *    <td>{@link #createControlFrame}</td>
 * </tr>
 * <tr>
 *    <td>CREATE DIALOG-BOX</td>
 *    <td>{@link #createDialogBox}</td>
 * </tr>
 * <tr>
 *    <td>CREATE EDITOR</td>
 *    <td>{@link #createEditor}</td>
 * </tr>
 * <tr>
 *    <td>CREATE FILL-IN</td>
 *    <td>{@link #createFillIn}</td>
 * </tr>
 * <tr>
 *    <td>CREATE FRAME</td>
 *    <td>{@link #createFrame}</td>
 * </tr>
 * <tr>
 *    <td>CREATE IMAGE</td>
 *    <td>{@link #createImage}</td>
 * </tr>
 * <tr>
 *    <td>CREATE MENU</td>
 *    <td>{@link #createMenu}</td>
 * </tr>
 * <tr>
 *    <td>CREATE MENU-ITEM</td>
 *    <td>{@link #createMenuItem}</td>
 * </tr>
 * <tr>
 *    <td>CREATE RADIO-SET</td>
 *    <td>{@link #createRadioSet}</td>
 * </tr>
 * <tr>
 *    <td>CREATE RECTANGLE</td>
 *    <td>{@link #createRectangle}</td>
 * </tr>
 * <tr>
 *    <td>CREATE SELECTION-LIST</td>
 *    <td>{@link #createSelectionList}</td>
 * </tr>
 * <tr>
 *    <td>CREATE SLIDER</td>
 *    <td>{@link #createSlider}</td>
 * </tr>
 * <tr>
 *    <td>CREATE SUB-MENU</td>
 *    <td>{@link #createSubMenu}</td>
 * </tr>
 * <tr>
 *    <td>CREATE TEXT</td>
 *    <td>{@link #createText}</td>
 * </tr>
 * <tr>
 *    <td>CREATE TOGGLE-BOX</td>
 *    <td>{@link #createToggleBox}</td>
 * </tr>
 * <tr>
 *    <td>CREATE WINDOW</td>
 *    <td>{@link #createWindow}</td>
 * </tr>
 * </table>
 * 
 * TODO: once the widget has been created, its attributes can be set regardless if is attached to
 * a frame or not. Currently, P2J allows the attributes to be set only if the widget is attached
 * to a frame, as the widget is not valid until is attached to a frame.
 */
public class DynamicWidgetFactory
{
   /** A map of widget names to their widget classes. */
   private static final Map<String, Class<? extends GenericWidget>> widgetFactory = 
      new HashMap<String, Class<? extends GenericWidget>>();
   
   static
   {
      widgetFactory.put(LegacyResource.BUTTON,         ButtonWidget.class);
      widgetFactory.put(LegacyResource.COMBO_BOX,      ComboBoxWidget.class);
      widgetFactory.put(LegacyResource.CONTROL_FRAME,  null); // TODO: what is the widget ?
      widgetFactory.put(LegacyResource.DIALOG_BOX,     null); // TODO: what is the widget ?
      widgetFactory.put(LegacyResource.EDITOR,         EditorWidget.class);
      widgetFactory.put(LegacyResource.FILL_IN,        FillInWidget.class);
      widgetFactory.put(LegacyResource.FRAME,          FrameWidget.class);
      widgetFactory.put(LegacyResource.IMAGE,          null); // TODO: what is the widget ?
      widgetFactory.put(LegacyResource.MENU,           MenuWidget.class); 
      widgetFactory.put(LegacyResource.MENU_ITEM,      null); // TODO: what is the widget ?
      widgetFactory.put(LegacyResource.RADIO_SET,      RadioSetWidget.class);
      widgetFactory.put(LegacyResource.RECTANGLE,      RectangleWidget.class);
      widgetFactory.put(LegacyResource.SELECTION_LIST, SelectionListWidget.class);
      widgetFactory.put(LegacyResource.SLIDER,         null); // TODO: what is the widget ?
      widgetFactory.put(LegacyResource.SUB_MENU,       null); // TODO: what is the widget ?
      widgetFactory.put(LegacyResource.TEXT,           TextWidget.class); // TODO: what is the widget ?
      widgetFactory.put(LegacyResource.TOGGLE_BOX,     ToggleBoxWidget.class);
      widgetFactory.put(LegacyResource.WINDOW,         WindowWidget.class);
   }
   
   /**
    * Create a widget, using the given widget type. This must be a value from the following list
    * (case insensitive): BUTTON, COMBO-BOX, CONTROL-FRAME, DIALOG-BOX, EDITOR, FILL-IN, FRAME,
    * IMAGE, MENU, MENU-ITEM, RADIO-SET, RECTANGLE, SELECTION-LIST, SLIDER, SUB-MENU, TEXT,
    * TOGGLE-BOX, WINDOW.
    * <p>
    * If the <code>widgetType</code> is one of the above strings, the appropriate factory API to
    * create the widget will be called.
    * <p>
    * If available, the widget will be saved in the unnamed widget pool.
    *  
    * @param    widgetType
    *           The widget type. Must be not-null.
    * @param    h
    *           A handle where to save the create widget. Must be not-null.
    */
   public static void create(String widgetType, handle h)
   {
      create(widgetType, h, (character) null);
   }

   /**
    * Create a widget, using the given widget type. This must be a value from the following list
    * (case insensitive): BUTTON, COMBO-BOX, CONTROL-FRAME, DIALOG-BOX, EDITOR, FILL-IN, FRAME,
    * IMAGE, MENU, MENU-ITEM, RADIO-SET, RECTANGLE, SELECTION-LIST, SLIDER, SUB-MENU, TEXT,
    * TOGGLE-BOX, WINDOW.
    * <p>
    * If the <code>widgetType</code> is one of the above strings, the appropriate factory API to
    * create the widget will be called.
    * <p>
    * If available, the widget will be saved in the unnamed widget pool.
    *  
    * @param    widgetType
    *           The widget type. Must be not-null.
    * @param    h
    *           A handle where to save the create widget. Must be not-null.
    * @param    widgetPool
    *           The named widget pool where to save the widget. If <code>null</code>, the unnamed
    *           widget pool will be used.
    */
   public static void create(String widgetType, handle h, String widgetPool)
   {
      create(widgetType, h, (widgetPool == null ? null : new character(widgetPool)));
   }
   
   /**
    * Create a widget, using the given widget type. This must be a value from the following list
    * (case insensitive): BUTTON, COMBO-BOX, CONTROL-FRAME, DIALOG-BOX, EDITOR, FILL-IN, FRAME,
    * IMAGE, MENU, MENU-ITEM, RADIO-SET, RECTANGLE, SELECTION-LIST, SLIDER, SUB-MENU, TEXT,
    * TOGGLE-BOX, WINDOW.
    * <p>
    * If the <code>widgetType</code> is one of the above strings, the appropriate factory API to
    * create the widget will be called.
    * <p>
    * If available, the widget will be saved in the unnamed widget pool.
    *  
    * @param    widgetType
    *           The widget type. Must be not-null.
    * @param    h
    *           A handle where to save the create widget. Must be not-null.
    * @param    widgetPool
    *           The named widget pool where to save the widget. If <code>null</code>, the unnamed
    *           widget pool will be used.
    */
   public static void create(String widgetType, handle h, character widgetPool)
   {
      // validate widget pool before anything else!!!
      if (!WidgetPool.validWidgetPool(widgetPool))
      {
         return;
      }
      
      Class<? extends GenericWidget> widCls = widgetFactory.get(widgetType.toUpperCase());
      
      if (widCls == null)
      {
         String msg = String.format("Invalid widget type %s", widgetType);
         ErrorManager.recordOrThrowError(3180, msg, false);
         
         // if the widget type is not known, do not touch the handle.
         return;
      }

      if (LegacyResource.WINDOW.equals(widgetType) && !isAbleToCreateWindow())
      {
         if (!EnvironmentOps.isBatchMode().booleanValue())
         {
            h.assign(LogicalTerminal.currentWindow());
         }  
         return;
      }

      try
      {
         Constructor<? extends GenericWidget> ctor = widCls.getDeclaredConstructor(boolean.class);
         WrappedResource res = ctor.newInstance(true);

         h.assign(res);
         
         // assign ID for new widget 
         int newId = WidgetId.nextID().getId();
         h.unwrapWidget().setId(newId);
         
         WidgetPool.addResource(res, widgetPool);
      }
      catch (Exception e)
      {
         // any exception here converts in a runtime exception
         throw new RuntimeException("Could not create widget of type " + widgetType, e);
      }
   }

   /**
    * Create a BUTTON widget and save it in the given handle and in the unnamed pool, if 
    * available.
    * 
    * @param    h
    *           A handle where to save the create widget. Must be not-null.
    */
   public static void createButton(handle h)
   {
      createButton(h, (character) null);
   }

   /**
    * Create a BUTTON widget and save it in the given handle and in the specified pool 
    * (which can be the unnamed pool if <code>widgetPool</code> is null).
    * 
    * @param    h
    *           A handle where to save the create widget. Must be not-null.
    * @param    widgetPool
    *           The named widget pool where to save the widget. If <code>null</code>, the unnamed
    *           widget pool will be used.
    */
   public static void createButton(handle h, String widgetPool)
   {
      create(LegacyResource.BUTTON, h, widgetPool);
   }

   /**
    * Create a BUTTON widget and save it in the given handle and in the specified pool 
    * (which can be the unnamed pool if <code>widgetPool</code> is null).
    * 
    * @param    h
    *           A handle where to save the create widget. Must be not-null.
    * @param    widgetPool
    *           The named widget pool where to save the widget. If <code>null</code>, the unnamed
    *           widget pool will be used.
    */
   public static void createButton(handle h, character widgetPool)
   {
      create(LegacyResource.BUTTON, h, widgetPool);
   }

   /**
    * Create a COMBO-BOX widget and save it in the given handle and in the unnamed pool, if 
    * available.
    * 
    * @param    h
    *           A handle where to save the create widget. Must be not-null.
    */
   public static void createComboBox(handle h)
   {
      createComboBox(h, (character) null);
   }

   /**
    * Create a COMBO-BOX widget and save it in the given handle and in the specified pool 
    * (which can be the unnamed pool if <code>widgetPool</code> is null).
    * 
    * @param    h
    *           A handle where to save the create widget. Must be not-null.
    * @param    widgetPool
    *           The named widget pool where to save the widget. If <code>null</code>, the unnamed
    *           widget pool will be used.
    */
   public static void createComboBox(handle h, String widgetPool)
   {
      create(LegacyResource.COMBO_BOX, h, widgetPool);
   }

   /**
    * Create a COMBO-BOX widget and save it in the given handle and in the specified pool 
    * (which can be the unnamed pool if <code>widgetPool</code> is null).
    * 
    * @param    h
    *           A handle where to save the create widget. Must be not-null.
    * @param    widgetPool
    *           The named widget pool where to save the widget. If <code>null</code>, the unnamed
    *           widget pool will be used.
    */
   public static void createComboBox(handle h, character widgetPool)
   {
      create(LegacyResource.COMBO_BOX, h, widgetPool);
   }

   /**
    * Create a CONTROL-FRAME widget and save it in the given handle and in the unnamed pool, if 
    * available.
    * 
    * @param    h
    *           A handle where to save the create widget. Must be not-null.
    */
   public static void createControlFrame(handle h)
   {
      createControlFrame(h, (character) null);
   }

   /**
    * Create a CONTROL-FRAME widget and save it in the given handle and in the specified pool 
    * (which can be the unnamed pool if <code>widgetPool</code> is null).
    * 
    * @param    h
    *           A handle where to save the create widget. Must be not-null.
    * @param    widgetPool
    *           The named widget pool where to save the widget. If <code>null</code>, the unnamed
    *           widget pool will be used.
    */
   public static void createControlFrame(handle h, String widgetPool)
   {
      create(LegacyResource.CONTROL_FRAME, h, widgetPool);
   }

   /**
    * Create a CONTROL-FRAME widget and save it in the given handle and in the specified pool 
    * (which can be the unnamed pool if <code>widgetPool</code> is null).
    * 
    * @param    h
    *           A handle where to save the create widget. Must be not-null.
    * @param    widgetPool
    *           The named widget pool where to save the widget. If <code>null</code>, the unnamed
    *           widget pool will be used.
    */
   public static void createControlFrame(handle h, character widgetPool)
   {
      create(LegacyResource.CONTROL_FRAME, h, widgetPool);
   }

   /**
    * Create a DIALOG-BOX widget and save it in the given handle and in the unnamed pool, if 
    * available.
    * 
    * @param    h
    *           A handle where to save the create widget. Must be not-null.
    */
   public static void createDialogBox(handle h)
   {
      createDialogBox(h, (character) null);
   }

   /**
    * Create a DIALOG-BOX widget and save it in the given handle and in the specified pool 
    * (which can be the unnamed pool if <code>widgetPool</code> is null).
    * 
    * @param    h
    *           A handle where to save the create widget. Must be not-null.
    * @param    widgetPool
    *           The named widget pool where to save the widget. If <code>null</code>, the unnamed
    *           widget pool will be used.
    */
   public static void createDialogBox(handle h, String widgetPool)
   {
      create(LegacyResource.DIALOG_BOX, h, widgetPool);
   }

   /**
    * Create a DIALOG-BOX widget and save it in the given handle and in the specified pool 
    * (which can be the unnamed pool if <code>widgetPool</code> is null).
    * 
    * @param    h
    *           A handle where to save the create widget. Must be not-null.
    * @param    widgetPool
    *           The named widget pool where to save the widget. If <code>null</code>, the unnamed
    *           widget pool will be used.
    */
   public static void createDialogBox(handle h, character widgetPool)
   {
      create(LegacyResource.DIALOG_BOX, h, widgetPool);
   }

   /**
    * Create a EDITOR widget and save it in the given handle and in the unnamed pool, if 
    * available.
    * 
    * @param    h
    *           A handle where to save the create widget. Must be not-null.
    */
   public static void createEditor(handle h)
   {
      createEditor(h, (character) null);
   }

   /**
    * Create a EDITOR widget and save it in the given handle and in the specified pool 
    * (which can be the unnamed pool if <code>widgetPool</code> is null).
    * 
    * @param    h
    *           A handle where to save the create widget. Must be not-null.
    * @param    widgetPool
    *           The named widget pool where to save the widget. If <code>null</code>, the unnamed
    *           widget pool will be used.
    */
   public static void createEditor(handle h, String widgetPool)
   {
      create(LegacyResource.EDITOR, h, widgetPool);
   }

   /**
    * Create a EDITOR widget and save it in the given handle and in the specified pool 
    * (which can be the unnamed pool if <code>widgetPool</code> is null).
    * 
    * @param    h
    *           A handle where to save the create widget. Must be not-null.
    * @param    widgetPool
    *           The named widget pool where to save the widget. If <code>null</code>, the unnamed
    *           widget pool will be used.
    */
   public static void createEditor(handle h, character widgetPool)
   {
      create(LegacyResource.EDITOR, h, widgetPool);
   }

   /**
    * Create a FILL-IN widget and save it in the given handle and in the unnamed pool, if 
    * available.
    * 
    * @param    h
    *           A handle where to save the create widget. Must be not-null.
    */
   public static void createFillIn(handle h)
   {
      createFillIn(h, (character) null);
   }

   /**
    * Create a FILL-IN widget and save it in the given handle and in the specified pool 
    * (which can be the unnamed pool if <code>widgetPool</code> is null).
    * 
    * @param    h
    *           A handle where to save the create widget. Must be not-null.
    * @param    widgetPool
    *           The named widget pool where to save the widget. If <code>null</code>, the unnamed
    *           widget pool will be used.
    */
   public static void createFillIn(handle h, String widgetPool)
   {
      create(LegacyResource.FILL_IN, h, widgetPool);
   }

   /**
    * Create a FILL-IN widget and save it in the given handle and in the specified pool 
    * (which can be the unnamed pool if <code>widgetPool</code> is null).
    * 
    * @param    h
    *           A handle where to save the create widget. Must be not-null.
    * @param    widgetPool
    *           The named widget pool where to save the widget. If <code>null</code>, the unnamed
    *           widget pool will be used.
    */
   public static void createFillIn(handle h, character widgetPool)
   {
      create(LegacyResource.FILL_IN, h, widgetPool);
   }

   /**
    * Create a FRAME widget and save it in the given handle and in the unnamed pool, if 
    * available.
    * 
    * @param    h
    *           A handle where to save the create widget. Must be not-null.
    */
   public static void createFrame(handle h)
   {
      createFrame(h, (character) null);
   }

   /**
    * Create a FRAME widget and save it in the given handle and in the specified pool 
    * (which can be the unnamed pool if <code>widgetPool</code> is null).
    * 
    * @param    h
    *           A handle where to save the create widget. Must be not-null.
    * @param    widgetPool
    *           The named widget pool where to save the widget. If <code>null</code>, the unnamed
    *           widget pool will be used.
    */
   public static void createFrame(handle h, String widgetPool)
   {
      create(LegacyResource.FRAME, h, widgetPool);
   }

   /**
    * Create a FRAME widget and save it in the given handle and in the specified pool 
    * (which can be the unnamed pool if <code>widgetPool</code> is null).
    * 
    * @param    h
    *           A handle where to save the create widget. Must be not-null.
    * @param    widgetPool
    *           The named widget pool where to save the widget. If <code>null</code>, the unnamed
    *           widget pool will be used.
    */
   public static void createFrame(handle h, character widgetPool)
   {
      create(LegacyResource.FRAME, h, widgetPool);
   }

   /**
    * Create a IMAGE widget and save it in the given handle and in the unnamed pool, if 
    * available.
    * 
    * @param    h
    *           A handle where to save the create widget. Must be not-null.
    */
   public static void createImage(handle h)
   {
      createImage(h, (character) null);
   }

   /**
    * Create a IMAGE widget and save it in the given handle and in the specified pool 
    * (which can be the unnamed pool if <code>widgetPool</code> is null).
    * 
    * @param    h
    *           A handle where to save the create widget. Must be not-null.
    * @param    widgetPool
    *           The named widget pool where to save the widget. If <code>null</code>, the unnamed
    *           widget pool will be used.
    */
   public static void createImage(handle h, String widgetPool)
   {
      create(LegacyResource.IMAGE, h, widgetPool);
   }

   /**
    * Create a IMAGE widget and save it in the given handle and in the specified pool 
    * (which can be the unnamed pool if <code>widgetPool</code> is null).
    * 
    * @param    h
    *           A handle where to save the create widget. Must be not-null.
    * @param    widgetPool
    *           The named widget pool where to save the widget. If <code>null</code>, the unnamed
    *           widget pool will be used.
    */
   public static void createImage(handle h, character widgetPool)
   {
      create(LegacyResource.IMAGE, h, widgetPool);
   }

   /**
    * Create a MENU widget and save it in the given handle and in the unnamed pool, if 
    * available.
    * 
    * @param    h
    *           A handle where to save the create widget. Must be not-null.
    */
   public static void createMenu(handle h)
   {
      createMenu(h, (character) null);
   }

   /**
    * Create a MENU widget and save it in the given handle and in the specified pool 
    * (which can be the unnamed pool if <code>widgetPool</code> is null).
    * 
    * @param    h
    *           A handle where to save the create widget. Must be not-null.
    * @param    widgetPool
    *           The named widget pool where to save the widget. If <code>null</code>, the unnamed
    *           widget pool will be used.
    */
   public static void createMenu(handle h, String widgetPool)
   {
      create(LegacyResource.MENU, h, widgetPool);
   }

   /**
    * Create a MENU widget and save it in the given handle and in the specified pool 
    * (which can be the unnamed pool if <code>widgetPool</code> is null).
    * 
    * @param    h
    *           A handle where to save the create widget. Must be not-null.
    * @param    widgetPool
    *           The named widget pool where to save the widget. If <code>null</code>, the unnamed
    *           widget pool will be used.
    */
   public static void createMenu(handle h, character widgetPool)
   {
      create(LegacyResource.MENU, h, widgetPool);
   }

   /**
    * Create a MENU-ITEM widget and save it in the given handle and in the unnamed pool, if 
    * available.
    * 
    * @param    h
    *           A handle where to save the create widget. Must be not-null.
    */
   public static void createMenuItem(handle h)
   {
      createMenuItem(h, (character) null);
   }

   /**
    * Create a MENU-ITEM widget and save it in the given handle and in the specified pool 
    * (which can be the unnamed pool if <code>widgetPool</code> is null).
    * 
    * @param    h
    *           A handle where to save the create widget. Must be not-null.
    * @param    widgetPool
    *           The named widget pool where to save the widget. If <code>null</code>, the unnamed
    *           widget pool will be used.
    */
   public static void createMenuItem(handle h, String widgetPool)
   {
      create(LegacyResource.MENU_ITEM, h, widgetPool);
   }

   /**
    * Create a MENU-ITEM widget and save it in the given handle and in the specified pool 
    * (which can be the unnamed pool if <code>widgetPool</code> is null).
    * 
    * @param    h
    *           A handle where to save the create widget. Must be not-null.
    * @param    widgetPool
    *           The named widget pool where to save the widget. If <code>null</code>, the unnamed
    *           widget pool will be used.
    */
   public static void createMenuItem(handle h, character widgetPool)
   {
      create(LegacyResource.MENU_ITEM, h, widgetPool);
   }

   /**
    * Create a RADIO-SET widget and save it in the given handle and in the unnamed pool, if 
    * available.
    * 
    * @param    h
    *           A handle where to save the create widget. Must be not-null.
    */
   public static void createRadioSet(handle h)
   {
      createRadioSet(h, (character) null);
   }

   /**
    * Create a RADIO-SET widget and save it in the given handle and in the specified pool 
    * (which can be the unnamed pool if <code>widgetPool</code> is null).
    * 
    * @param    h
    *           A handle where to save the create widget. Must be not-null.
    * @param    widgetPool
    *           The named widget pool where to save the widget. If <code>null</code>, the unnamed
    *           widget pool will be used.
    */
   public static void createRadioSet(handle h, String widgetPool)
   {
      create(LegacyResource.RADIO_SET, h, widgetPool);
   }

   /**
    * Create a RADIO-SET widget and save it in the given handle and in the specified pool 
    * (which can be the unnamed pool if <code>widgetPool</code> is null).
    * 
    * @param    h
    *           A handle where to save the create widget. Must be not-null.
    * @param    widgetPool
    *           The named widget pool where to save the widget. If <code>null</code>, the unnamed
    *           widget pool will be used.
    */
   public static void createRadioSet(handle h, character widgetPool)
   {
      create(LegacyResource.RADIO_SET, h, widgetPool);
   }

   /**
    * Create a RECTANGLE widget and save it in the given handle and in the unnamed pool, if 
    * available.
    * 
    * @param    h
    *           A handle where to save the create widget. Must be not-null.
    */
   public static void createRectangle(handle h)
   {
      createRectangle(h, (character) null);
   }

   /**
    * Create a RECTANGLE widget and save it in the given handle and in the specified pool 
    * (which can be the unnamed pool if <code>widgetPool</code> is null).
    * 
    * @param    h
    *           A handle where to save the create widget. Must be not-null.
    * @param    widgetPool
    *           The named widget pool where to save the widget. If <code>null</code>, the unnamed
    *           widget pool will be used.
    */
   public static void createRectangle(handle h, String widgetPool)
   {
      create(LegacyResource.RECTANGLE, h, widgetPool);
   }

   /**
    * Create a RECTANGLE widget and save it in the given handle and in the specified pool 
    * (which can be the unnamed pool if <code>widgetPool</code> is null).
    * 
    * @param    h
    *           A handle where to save the create widget. Must be not-null.
    * @param    widgetPool
    *           The named widget pool where to save the widget. If <code>null</code>, the unnamed
    *           widget pool will be used.
    */
   public static void createRectangle(handle h, character widgetPool)
   {
      create(LegacyResource.RECTANGLE, h, widgetPool);
   }

   /**
    * Create a SELECTION-LIST widget and save it in the given handle and in the unnamed pool, if 
    * available.
    * 
    * @param    h
    *           A handle where to save the create widget. Must be not-null.
    */
   public static void createSelectionList(handle h)
   {
      createSelectionList(h, (character) null);
   }

   /**
    * Create a SELECTION-LIST widget and save it in the given handle and in the specified pool 
    * (which can be the unnamed pool if <code>widgetPool</code> is null).
    * 
    * @param    h
    *           A handle where to save the create widget. Must be not-null.
    * @param    widgetPool
    *           The named widget pool where to save the widget. If <code>null</code>, the unnamed
    *           widget pool will be used.
    */
   public static void createSelectionList(handle h, String widgetPool)
   {
      create(LegacyResource.SELECTION_LIST, h, widgetPool);
   }

   /**
    * Create a SELECTION-LIST widget and save it in the given handle and in the specified pool 
    * (which can be the unnamed pool if <code>widgetPool</code> is null).
    * 
    * @param    h
    *           A handle where to save the create widget. Must be not-null.
    * @param    widgetPool
    *           The named widget pool where to save the widget. If <code>null</code>, the unnamed
    *           widget pool will be used.
    */
   public static void createSelectionList(handle h, character widgetPool)
   {
      create(LegacyResource.SELECTION_LIST, h, widgetPool);
   }

   /**
    * Create a SLIDER widget and save it in the given handle and in the unnamed pool, if 
    * available.
    * 
    * @param    h
    *           A handle where to save the create widget. Must be not-null.
    */
   public static void createSlider(handle h)
   {
      createSlider(h, (character) null);
   }

   /**
    * Create a SLIDER widget and save it in the given handle and in the specified pool 
    * (which can be the unnamed pool if <code>widgetPool</code> is null).
    * 
    * @param    h
    *           A handle where to save the create widget. Must be not-null.
    * @param    widgetPool
    *           The named widget pool where to save the widget. If <code>null</code>, the unnamed
    *           widget pool will be used.
    */
   public static void createSlider(handle h, String widgetPool)
   {
      create(LegacyResource.SLIDER, h, widgetPool);
   }

   /**
    * Create a SLIDER widget and save it in the given handle and in the specified pool 
    * (which can be the unnamed pool if <code>widgetPool</code> is null).
    * 
    * @param    h
    *           A handle where to save the create widget. Must be not-null.
    * @param    widgetPool
    *           The named widget pool where to save the widget. If <code>null</code>, the unnamed
    *           widget pool will be used.
    */
   public static void createSlider(handle h, character widgetPool)
   {
      create(LegacyResource.SLIDER, h, widgetPool);
   }

   /**
    * Create a SUB-MENU widget and save it in the given handle and in the unnamed pool, if 
    * available.
    * 
    * @param    h
    *           A handle where to save the create widget. Must be not-null.
    */
   public static void createSubMenu(handle h)
   {
      createSubMenu(h, (character) null);
   }

   /**
    * Create a SUB-MENU widget and save it in the given handle and in the specified pool 
    * (which can be the unnamed pool if <code>widgetPool</code> is null).
    * 
    * @param    h
    *           A handle where to save the create widget. Must be not-null.
    * @param    widgetPool
    *           The named widget pool where to save the widget. If <code>null</code>, the unnamed
    *           widget pool will be used.
    */
   public static void createSubMenu(handle h, String widgetPool)
   {
      create(LegacyResource.SUB_MENU, h, widgetPool);
   }

   /**
    * Create a SUB-MENU widget and save it in the given handle and in the specified pool 
    * (which can be the unnamed pool if <code>widgetPool</code> is null).
    * 
    * @param    h
    *           A handle where to save the create widget. Must be not-null.
    * @param    widgetPool
    *           The named widget pool where to save the widget. If <code>null</code>, the unnamed
    *           widget pool will be used.
    */
   public static void createSubMenu(handle h, character widgetPool)
   {
      create(LegacyResource.SUB_MENU, h, widgetPool);
   }

   /**
    * Create a TEXT widget and save it in the given handle and in the unnamed pool, if 
    * available.
    * 
    * @param    h
    *           A handle where to save the create widget. Must be not-null.
    */
   public static void createText(handle h)
   {
      createText(h, (character) null);
   }

   /**
    * Create a TEXT widget and save it in the given handle and in the specified pool 
    * (which can be the unnamed pool if <code>widgetPool</code> is null).
    * 
    * @param    h
    *           A handle where to save the create widget. Must be not-null.
    * @param    widgetPool
    *           The named widget pool where to save the widget. If <code>null</code>, the unnamed
    *           widget pool will be used.
    */
   public static void createText(handle h, String widgetPool)
   {
      create(LegacyResource.TEXT, h, widgetPool);
   }

   /**
    * Create a TEXT widget and save it in the given handle and in the specified pool 
    * (which can be the unnamed pool if <code>widgetPool</code> is null).
    * 
    * @param    h
    *           A handle where to save the create widget. Must be not-null.
    * @param    widgetPool
    *           The named widget pool where to save the widget. If <code>null</code>, the unnamed
    *           widget pool will be used.
    */
   public static void createText(handle h, character widgetPool)
   {
      create(LegacyResource.TEXT, h, widgetPool);
   }

   /**
    * Create a TOGGLE-BOX widget and save it in the given handle and in the unnamed pool, if 
    * available.
    * 
    * @param    h
    *           A handle where to save the create widget. Must be not-null.
    */
   public static void createToggleBox(handle h)
   {
      createToggleBox(h, (character) null);
   }

   /**
    * Create a TOGGLE-BOX widget and save it in the given handle and in the specified pool 
    * (which can be the unnamed pool if <code>widgetPool</code> is null).
    * 
    * @param    h
    *           A handle where to save the create widget. Must be not-null.
    * @param    widgetPool
    *           The named widget pool where to save the widget. If <code>null</code>, the unnamed
    *           widget pool will be used.
    */
   public static void createToggleBox(handle h, String widgetPool)
   {
      create(LegacyResource.TOGGLE_BOX, h, widgetPool);
   }

   /**
    * Create a TOGGLE-BOX widget and save it in the given handle and in the specified pool 
    * (which can be the unnamed pool if <code>widgetPool</code> is null).
    * 
    * @param    h
    *           A handle where to save the create widget. Must be not-null.
    * @param    widgetPool
    *           The named widget pool where to save the widget. If <code>null</code>, the unnamed
    *           widget pool will be used.
    */
   public static void createToggleBox(handle h, character widgetPool)
   {
      create(LegacyResource.TOGGLE_BOX, h, widgetPool);
   }

   /**
    * Create a WINDOW widget and save it in the given handle and in the unnamed pool, if 
    * available.
    * 
    * @param    h
    *           A handle where to save the create widget. Must be not-null.
    */
   public static void createWindow(handle h)
   {
      createWindow(h, (character) null);
   }

   /**
    * Create a WINDOW widget and save it in the given handle and in the specified pool (which can
    * be the unnamed pool if <code>widgetPool</code> is null).
    * 
    * @param    h
    *           A handle where to save the create widget. Must be not-null.
    * @param    widgetPool
    *           The named widget pool where to save the widget. If <code>null</code>, the unnamed
    *           widget pool will be used.
    */
   public static void createWindow(handle h, String widgetPool)
   {
      create(LegacyResource.WINDOW, h, widgetPool);
   }

   /**
    * Create a WINDOW widget and save it in the given handle and in the specified pool (which can
    * be the unnamed pool if <code>widgetPool</code> is null).
    * 
    * @param    h
    *           A handle where to save the create widget. Must be not-null.
    * @param    widgetPool
    *           The named widget pool where to save the widget. If <code>null</code>, the unnamed
    *           widget pool will be used.
    */
   public static void createWindow(handle h, character widgetPool)
   {
      create(LegacyResource.WINDOW, h, widgetPool);
   }
   
   /**
    * Checks if it is possible to create new window. Character interfaces 
    * do not support new window creation.
    *
    * @return  <code>TRUE</code> - creation is possible, 
    *          <code>FALSE</code> - otherwise  
    */
   private static boolean isAbleToCreateWindow()
   {
      String msg = "CREATE WINDOW in character mode is not supported";
      boolean batchMode = EnvironmentOps.isBatchMode().booleanValue();
      if (LogicalTerminal.isChui() && batchMode)
      {
         int[] codes = new int[] { 4139, 3181 };
         String[] messages = new String[] 
                  { msg, "Cannot create widget of type WINDOW" };
         ErrorManager.recordOrThrowError(codes, messages, false, false);
         return false;
      }
      
      if (LogicalTerminal.isChui() && !batchMode)
      {
         ErrorManager.recordOrShowError(4139, msg, false, false, false);
         return false;
      }

      return true;
   }
}

